/*
 * Copyright 2007 Peter Hutterer
 * Copyright 2009 Przemysław Firszt
 *
 * Permission to use, copy, modify, distribute, and sell this software
 * and its documentation for any purpose is hereby granted without
 * fee, provided that the above copyright notice appear in all copies
 * and that both that copyright notice and this permission notice
 * appear in supporting documentation, and that the name of Red Hat
 * not be used in advertising or publicity pertaining to distribution
 * of the software without specific, written prior permission.  Red
 * Hat makes no representations about the suitability of this software
 * for any purpose.  It is provided "as is" without express or implied
 * warranty.
 *
 * THE AUTHORS DISCLAIM ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 * INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN
 * NO EVENT SHALL THE AUTHORS BE LIABLE FOR ANY SPECIAL, INDIRECT OR
 * CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS
 * OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT,
 * NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN
 * CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */
 
#ifdef HAVE_CONFIG_H
#include "config.h"
#endif


#include <linux/input.h>
#include <linux/types.h>
#include <unistd.h>

#include <xf86_OSproc.h>

#include <xf86.h>
#include <xf86Xinput.h>
#include <exevents.h>
#include <xorg-server.h>
#include <xorgVersion.h>
#include <xkbsrv.h>


#ifdef HAVE_PROPERTIES
#include <xserver-properties.h>
/* 1.6 has properties, but no labels */
#ifdef AXIS_LABEL_PROP
#define HAVE_LABELS
#else
#undef HAVE_LABELS
#endif

#endif


#include <stdio.h>
#include <sys/stat.h>
#include <sys/syscall.h>
#include <unistd.h>
#include <errno.h>
#include <sys/types.h>
#include <fcntl.h>
#include <xorgVersion.h>
#include <xf86Module.h>
#include <X11/Xatom.h>

#include "ts-input.h"
#include "nbus.h"

static void *inputThread(void *data);

static int TS_TouchScreenPreInit(InputDriverPtr drv, InputInfoPtr dev, int flags);
static void TS_TouchScreenUnInit(InputDriverPtr drv, InputInfoPtr pInfo, int flags);
static pointer TS_TouchScreenPlug(pointer module, pointer options, int *errmaj, int  *errmin);
static void TS_TouchScreenUnplug(pointer p);
static void TS_TouchScreenReadInput(InputInfoPtr pInfo);
static int TS_TouchScreenControl(DeviceIntPtr    device,int what);
static int _TS_TouchScreen_init_buttons(DeviceIntPtr device);
static int _TS_TouchScreen_init_axes(DeviceIntPtr device);


static unsigned short calib_x0, calib_dx, calib_y0, calib_dy;

#define NUM_READINGS 32 
static unsigned short dataStack[2][NUM_READINGS];
static int integrationCount;
static unsigned short previousAverage[2];


/* TS_TouchScreen_driver_name[] fixes a gcc warning:
 * "initialization discards 'const' qualifier from pointer target type" */
static char TS_TouchScreen_driver_name[] = "tstouch";

_X_EXPORT InputDriverRec TS_TouchScreen = {
    1,
    TS_TouchScreen_driver_name,
    NULL,
    TS_TouchScreenPreInit,
    TS_TouchScreenUnInit,
    NULL,
    0
};

static XF86ModuleVersionInfo TS_TouchScreenVersionRec =
{
    "tstouch",
    MODULEVENDORSTRING,
    MODINFOSTRING1,
    MODINFOSTRING2,
    XORG_VERSION_CURRENT,
    PACKAGE_VERSION_MAJOR, PACKAGE_VERSION_MINOR,
    PACKAGE_VERSION_PATCHLEVEL,
    ABI_CLASS_XINPUT,
    ABI_XINPUT_VERSION,
    MOD_CLASS_XINPUT,
    {0, 0, 0, 0}
};

_X_EXPORT XF86ModuleData tstouchModuleData =
{
    &TS_TouchScreenVersionRec,
    TS_TouchScreenPlug,
    TS_TouchScreenUnplug
};

static void
TS_TouchScreenUnplug(pointer p)
{
}

static pointer
TS_TouchScreenPlug(pointer        module,
           pointer        options,
           int            *errmaj,
           int            *errmin)
{
    xf86AddInputDriver(&TS_TouchScreen, module, 0);
    return module;
}


static void TS_TouchScreenGetCalibration(void) {
 
   /** TODO
      Make this function read the calibration values from
      wherever they are being stored.
    */
    
   calib_x0 = 101;
   calib_dx = 3904;
   
   calib_y0 = 175;
   calib_dy = 3891;    
}

static int TS_TouchScreenPreInit(InputDriverPtr  drv,
  InputInfoPtr pInfo, int flags)
{   
   int status;
   char *s;    
   int rc = BadAlloc;
   TS_TouchScreenDevicePtr    pTS_TouchScreen;

   pTS_TouchScreen = calloc(1, sizeof(TS_TouchScreenDeviceRec));
   if (!pTS_TouchScreen) {
      pInfo->private = NULL;
      xf86DeleteInput(pInfo, 0);
      goto error;
   }

   pInfo->private = pTS_TouchScreen;
    
   pInfo->flags = 0;
   pInfo->type_name = strdup(XI_MOUSE); /* see XI.h */
   pInfo->read_input = TS_TouchScreenReadInput; /* new data avl */
   pInfo->switch_mode = NULL; /* toggle absolute/relative mode */
   pInfo->device_control = TS_TouchScreenControl; /* enable/disable dev */
    
   /* process driver specific options */
    pTS_TouchScreen->device = xf86SetStrOption(pInfo->options,
                                       "Device",
                                       "/tmp/.ts-fifo");

    xf86Msg(X_INFO, "%s: Using device %s.\n", pInfo->name, pTS_TouchScreen->device);

    unlink(pTS_TouchScreen->device);
    status = mkfifo(pTS_TouchScreen->device, 0666);
    
    if (status) {
       if (status != EEXIST) {
          xf86Msg(X_ERROR, "%s: failed to create FIFO %s. err: %d",
                pInfo->name, pTS_TouchScreen->device, status);
          pInfo->private = NULL;
          free(pTS_TouchScreen);
          xf86DeleteInput(pInfo, 0);  
          goto error;
       }       
    }
               
    /* process generic options */
    xf86CollectInputOptions(pInfo, NULL);
    xf86ProcessCommonOptions(pInfo, pInfo->options);
        
    if ((s = xf86SetStrOption(pInfo->options, "MinX", NULL)))             
       pTS_TouchScreen->minX = strtol(s, NULL, 0);
    else 
       pTS_TouchScreen->minX = 0;
    
    if ((s = xf86SetStrOption(pInfo->options, "MaxX", NULL)))             
       pTS_TouchScreen->maxX = strtol(s, NULL, 0);
    else 
       pTS_TouchScreen->maxX = 4095;
                 
    if ((s = xf86SetStrOption(pInfo->options, "MinY", NULL)))             
       pTS_TouchScreen->minY = strtol(s, NULL, 0);
    else 
       pTS_TouchScreen->minY = 0;
    
    if ((s = xf86SetStrOption(pInfo->options, "MaxY", NULL)))             
       pTS_TouchScreen->maxY = strtol(s, NULL, 0);
    else 
       pTS_TouchScreen->maxY = 4095;
                     
    if ((s = xf86SetStrOption(pInfo->options, "IntegrationCount", NULL))) {             
        integrationCount = strtol(s, NULL, 0);
        if (integrationCount < 0 || integrationCount >= NUM_READINGS) {
            xf86Msg(X_ERROR, "%s: IntegrationCount out of range (0..%d)\n", 
               pInfo->name, NUM_READINGS-1);
            integrationCount = 0;
        }
    }
    else 
       integrationCount = 0;
    
    /* Assuming we have just one screen for now ... */
    
    pTS_TouchScreen->screenWidth = screenInfo.screens[0]->width;
    pTS_TouchScreen->screenHeight = screenInfo.screens[0]->height;
         
    /* Open sockets, init device files, etc. */
        
    pTS_TouchScreen->run = 0;
    pTS_TouchScreen->done = 0;
  
    if (pthread_create(&pTS_TouchScreen->thread, NULL, inputThread, pInfo)) {
      xf86Msg(X_ERROR, "%s: Failed to create thread.\n", pInfo->name);
 
       unlink(pTS_TouchScreen->device);
       pInfo->private = NULL;
       free(pTS_TouchScreen);
       xf86DeleteInput(pInfo, 0);  
      
      goto error;
   } 
           
   pInfo->fd = -1;
    
   return Success;
error:
   return rc;
}

static void TS_TouchScreenUnInit(InputDriverPtr drv,
                         InputInfoPtr   pInfo,
                         int            flags)
{
    TS_TouchScreenDevicePtr     pTS_TouchScreen = pInfo->private;
    if (pTS_TouchScreen->device)
    {
       
       pTS_TouchScreen->done = 1;
       sleep(1);  /* wait for thread to shut itself down */           
       unlink(pTS_TouchScreen->device);
           
       free(pTS_TouchScreen->device);
       pTS_TouchScreen->device = NULL;
       /* Common error - pInfo->private must be NULL or valid memoy before
        * passing into xf86DeleteInput */
        pInfo->private = NULL;
    }
        
    xf86DeleteInput(pInfo, 0);
}


static int
_TS_TouchScreen_init_buttons(DeviceIntPtr device)
{
    InputInfoPtr        pInfo = device->public.devicePrivate;
    TS_TouchScreenDevicePtr     pTS_TouchScreen = pInfo->private;
    CARD8               *map;
    int                 i;
    int                 ret = Success;
    const int           num_buttons = 2;

    map = calloc(num_buttons, sizeof(CARD8));

    for (i = 0; i < num_buttons; i++)
        map[i] = i;

    pTS_TouchScreen->labels = calloc(1, sizeof(Atom));

    if (!InitButtonClassDeviceStruct(device, num_buttons, pTS_TouchScreen->labels, map)) {
            xf86Msg(X_ERROR, "%s: Failed to register buttons.\n", pInfo->name);
            ret = BadAlloc;
    }

    free(map);
    return ret;
}

static void TS_TouchScreenInitAxesLabels(TS_TouchScreenDevicePtr pTS_TouchScreen, int natoms, Atom *atoms)
{
#ifdef HAVE_LABELS
    Atom atom;
    int axis;
    char **labels;
    int labels_len = 0;
    char *misc_label;

    labels     = rel_labels;
    labels_len = ArrayLength(rel_labels);
    misc_label = AXIS_LABEL_PROP_REL_MISC;

    memset(atoms, 0, natoms * sizeof(Atom));

    /* Now fill the ones we know */
    for (axis = 0; axis < labels_len; axis++)
    {
        if (pTS_TouchScreen->axis_map[axis] == -1)
            continue;

        atom = XIGetKnownProperty(labels[axis]);
        if (!atom) /* Should not happen */
            continue;

        atoms[pTS_TouchScreen->axis_map[axis]] = atom;
    }
#endif
}


static int
_TS_TouchScreen_init_axes(DeviceIntPtr device)
{
    InputInfoPtr        pInfo = device->public.devicePrivate;
    TS_TouchScreenDevicePtr     pTS_TouchScreen = pInfo->private;
    int                 i;
    const int           num_axes = 2;
    Atom                * atoms;

    pTS_TouchScreen->num_vals = num_axes;
    atoms = calloc(pTS_TouchScreen->num_vals, sizeof(Atom));

    TS_TouchScreenInitAxesLabels(pTS_TouchScreen, pTS_TouchScreen->num_vals, atoms);
    if (!InitValuatorClassDeviceStruct(device,
                num_axes,
#if GET_ABI_MAJOR(ABI_XINPUT_VERSION) >= 7
                atoms,
#endif	
                GetMotionHistorySize(),
                0))
        return BadAlloc;

    for (i = 0; i < pTS_TouchScreen->axes; i++) {
            xf86InitValuatorAxisStruct(device, i, *pTS_TouchScreen->labels,
                                       -1, -1, 1, 1, 1, 1);
            xf86InitValuatorDefaults(device, i);
    }
    free(atoms);
    return Success;
}

static int TS_TouchScreenControl(DeviceIntPtr    device,
                         int             what)
{
    InputInfoPtr  pInfo = device->public.devicePrivate;
    TS_TouchScreenDevicePtr pTS_TouchScreen = pInfo->private;

    switch(what)
    {
        case DEVICE_INIT:
            _TS_TouchScreen_init_buttons(device);
            _TS_TouchScreen_init_axes(device);
            break;

        /* Switch device on.  Establish socket, start event delivery.  */
        case DEVICE_ON:
            xf86Msg(X_INFO, "%s: On.\n", pInfo->name);
            if (device->public.on)
                    break;
                        
            SYSCALL(pInfo->fd = open(pTS_TouchScreen->device, O_RDONLY));
            if (pInfo->fd < 0)
            {
                xf86Msg(X_ERROR, "%s: cannot open device '%s'\n", 
                   pInfo->name, pTS_TouchScreen->device);
                return BadRequest;
            }

            pTS_TouchScreen->run = 1;
            xf86FlushInput(pInfo->fd);
            xf86AddEnabledDevice(pInfo);
            device->public.on = TRUE;
            break;
       case DEVICE_OFF:
            xf86Msg(X_INFO, "%s: Off.\n", pInfo->name);
            if (!device->public.on)
                break;
                        
            xf86RemoveEnabledDevice(pInfo);
            close(pInfo->fd);
            pInfo->fd = -1; 
             
            pTS_TouchScreen->run = 0; 
            device->public.on = FALSE;
            break;
      case DEVICE_CLOSE:
            /* free what we have to free */
            break;
    }
        
    return Success;
}



static unsigned short GetAverageValue(int axis, unsigned short latestReading) {

   static int inited[2];
   static int index[2], i, j;
   unsigned long v;
   
   if (axis > 1)
      return 0;
   
   if (axis < 0) {
      inited[0] = 0;
      inited[1] = 0;
      return 0;      
   }
   
   j = integrationCount;
   
   if (!inited[axis]) {      
      for(i=0; i < j; i++)
         dataStack[axis][i] = latestReading;      
      inited[axis] = 1;        
   }
   
   dataStack[axis][index[axis]] = latestReading;
      
   for(v=i=0; i < j; i++)
      v += dataStack[axis][i];
      
   if (--index[axis] < 0)
      index[axis] = j-1;
                           
   return previousAverage[axis] = (v / j);   
}




static void *inputThread(void *data) {
    
   InputInfoPtr pInfo = data;
   TS_TouchScreenDevicePtr pTS_TouchScreen = pInfo->private;
   struct timespec ts;
   int fd;
   int bounce;
      
   TS_TouchScreenDeviceReport report;
   
   int x,y, screenWidth, screenHeight;
   int prev_x, prev_y;
   int press;

   TS_TouchScreenGetCalibration();
   
   screenWidth  = pTS_TouchScreen->screenWidth;
   screenHeight = pTS_TouchScreen->screenHeight; 
   
   prev_x = screenWidth / 2; 
   prev_y = screenHeight / 2;
   
   SYSCALL(fd = open(pTS_TouchScreen->device, O_WRONLY)); 
   
   press = 0;
   bounce = 3;   
   
   memset(&report, 0, sizeof(report));
   
   while(! pTS_TouchScreen->done) {
      
      if (report.event) {            
         write(fd, &report, sizeof(report));
         report.event = 0;
      }
      
      ts.tv_sec = 0;
      ts.tv_nsec = 10000000;  /* sleep 10ms (so sample rate == 100Hz) */
      nanosleep(&ts, NULL);
                  
      if (pTS_TouchScreen->run) {      

         nbuslock();         
         x = nbus_peek16(0x58);
         y = nbus_peek16(0x5A);                  
         nbusunlock();
                    
         if ((x & 1) || (y & 1)) {     /* Screen is touched */
                      
             x = ((~x) >> 4) & 0xfff;
             y = ((~y) >> 4) & 0xfff;
                
             //xf86Msg(X_INFO, "%d %d\r\n", x,y);

             if (integrationCount > 1) { 
                x = GetAverageValue(0, x);
                y = GetAverageValue(1, y);
             }
             
             x = (x-calib_x0) * screenWidth / (calib_dx - calib_x0);
             y = (y-calib_y0) * screenHeight / (calib_dy - calib_y0);
      
             //xf86Msg(X_INFO, "%d %d\r\n", x,y);
                      
             if (x != prev_x || y != prev_y) {
                                                   
                report.x = x;
                report.y = y;
                report.event |= CURSOR_MOVE_EVENT;
                                                         
                prev_x = x;
                prev_y = y;                                                      
             }
             
            if (press == 0) {               
               bounce--;
               if (bounce > 0)
                  continue;             
               
               press = 1;               
               bounce = 3;                                                
               report.event = BUTTON_DOWN_EVENT;                                             
            }    
             
         } else {    /* Screen NOT touched */
            
            if (press) {               
               bounce--;
               if (bounce > 0)
                  continue;               
               
               press = 0;
               bounce = 3;               
               report.event = BUTTON_UP_EVENT;

               if (integrationCount > 1)
                  GetAverageValue(-1, 0); /* reset the integrator */
               
            }            
         }                   
      }
   }

   SYSCALL(close(fd));
   
   return NULL;
}

static void TS_TouchScreenReadInput(InputInfoPtr pInfo)
{    
   TS_TouchScreenDeviceReport report;
    
   while(xf86WaitForInput(pInfo->fd, 0) > 0)
   {
       read(pInfo->fd, &report, sizeof(report));
               
       if (report.event & CURSOR_MOVE_EVENT) {
          
          xf86Msg(X_INFO, "%d %d\r\n", report.x, report.y);
          
          xf86PostMotionEvent(pInfo->dev,
                          1, /* is_absolute */
                          0, /* first_valuator */
                          2, /* num_valuators */
                          (int)report.x, (int)report.y);
       }
       
       if (report.event & BUTTON_DOWN_EVENT) {
          xf86Msg(X_INFO, "DOWN\n");
          xf86PostButtonEvent(pInfo->dev, 0, 1, 1, 0, 0);
       }

       if (report.event & BUTTON_UP_EVENT) {
          xf86Msg(X_INFO, "UP\n");
          xf86PostButtonEvent(pInfo->dev, 0, 1, 0, 0, 0);
       }        
   }
}
